from typing import Generic, TypeVar

from rainbowadn.core.hashpoint import HashPoint
from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.trees.binary.actions.binarytreeaction import BinaryTreeAction
from rainbowadn.data.collection.trees.binary.actions.symmetric import InnerOuter, OuterInner, Symmetric
from rainbowadn.data.collection.trees.binary.binarytreebalancingprotocol import BinaryTreeBalancingProtocol
from rainbowadn.data.collection.trees.binary.binarytreecase import BinaryTreeCase
from rainbowadn.data.collection.trees.binary.binarytreecreationprotocol import BinaryTreeCreationProtocol

__all__ = ('AVL',)

ActiveKeyType = TypeVar('ActiveKeyType')
TreeType = TypeVar('TreeType')


class AVL(BinaryTreeBalancingProtocol[ActiveKeyType, Integer, TreeType]):
    def empty_metadata(self) -> HashPoint[Integer]:
        return HashPoint.of(Integer(0))

    def metadata(
            self,
            treel: TreeType,
            treer: TreeType,
            key: HashPoint[ActiveKeyType],
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> HashPoint[Integer]:
        return HashPoint.of(
            Integer(1 + max(self.height(treel, protocol), self.height(treer, protocol)))
        )

    @classmethod
    def height(
            cls,
            tree: TreeType,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> int:
        return HeightAction().on(protocol, tree)

    def balance(
            self,
            tree: TreeType,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        return BalanceAction().on(protocol, tree)


class HeightAction(
    BinaryTreeAction[ActiveKeyType, Integer, TreeType, int],
    Generic[ActiveKeyType, TreeType]
):
    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType],
            tree: TreeType
    ) -> int:
        return 0

    def on_split(
            self,
            case: BinaryTreeCase[ActiveKeyType, Integer, TreeType]
    ) -> int:
        return case.split.metadata.resolve().integer


class BalanceAction(
    BinaryTreeAction[ActiveKeyType, Integer, TreeType, TreeType],
    Generic[ActiveKeyType, TreeType]
):
    def on_null(
            self,
            protocol: BinaryTreeCreationProtocol[ActiveKeyType, Integer, TreeType],
            tree: TreeType
    ) -> TreeType:
        return tree

    def on_split(
            self,
            case: BinaryTreeCase[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        split, protocol = case.split, case.protocol
        delta = AVL.height(split.treel, protocol) - AVL.height(split.treer, protocol)
        assert isinstance(delta, int)
        if delta < -1:
            return self.on_symmetric(InnerOuter(case))
        elif delta > 1:
            return self.on_symmetric(OuterInner(case))
        else:
            return case.tree

    @classmethod
    def on_symmetric(
            cls,
            symmetric: Symmetric[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        protocol, split = symmetric.case.protocol, symmetric.case.split
        splito = protocol.fsplit(symmetric.outer(split))
        if AVL.height(symmetric.inner(splito), protocol) > AVL.height(symmetric.outer(splito), protocol):
            splitoi = protocol.fsplit(symmetric.inner(splito))
            return symmetric.tree(
                symmetric.tree(symmetric.inner(split), symmetric.inner(splitoi), split.key),
                symmetric.tree(symmetric.outer(splitoi), symmetric.outer(splito), splito.key),
                splitoi.key
            )
        else:
            return symmetric.tree(
                symmetric.tree(symmetric.inner(split), symmetric.inner(splito), split.key),
                symmetric.outer(splito),
                splito.key
            )
